import random

from paramga.encoding import decode_parameters, encode_parameters


def param_crossover(param_a, param_b):
    encoded_param_a = encode_parameters(param_a)
    encoded_param_b = encode_parameters(param_b)
    number_of_params = len(encoded_param_a)
    k1 = random.randint(0, number_of_params-1)
    k2 = random.randint(0, number_of_params-1)

    pa = encoded_param_a[0:k1] if k1 < k2 else encoded_param_b[0:k2]
    pb = encoded_param_a[k1:k2] if k1 < k2 else encoded_param_b[k2: k1]
    pc = encoded_param_a[k2:] if k1 < k2 else encoded_param_b[k1:]

    new_encoded_param = pa + pb + pc
    new_param = decode_parameters(new_encoded_param)
    return new_param



"""Genetic Algorith approach to optimizing input model parameters.

Uses method described in "Using genetic algorithms to optimise model parameters, Q.J.Wang(1997).
"""

# %%
from matplotlib import pyplot as plt
import numpy
import random

example_param_base = {
    "foo": 10,
    "bar": 100,
}

ga_config = {
    "foo": {
        "key": "foo",
        "type": "number",
        "min": 3,
        "max": 8,
    }
}

data = [1, 2, 3, 3]


def encode_parameters(param):
    return list(param.items())


def decode_parameters(param):
    return {k: v for k, v in param}


def combine_param_state(param_a, param_b):
    encoded_param_a = encode_parameters(param_a)
    encoded_param_b = encode_parameters(param_b)
    number_of_params = len(encoded_param_a)
    k1 = random.randint(0, number_of_params-1)
    k2 = random.randint(0, number_of_params-1)

    pa = encoded_param_a[0:k1] if k1 < k2 else encoded_param_b[0:k2]
    pb = encoded_param_a[k1:k2] if k1 < k2 else encoded_param_b[k2: k1]
    pc = encoded_param_a[k2:] if k1 < k2 else encoded_param_b[k1:]

    new_encoded_param = pa + pb + pc
    new_param = decode_parameters(new_encoded_param)
    return new_param


def generate_param_state(param_base, gaconf):
    new_param = {**param_base, 'foo': random.randint(0, 10)}
    return new_param


def mutate_param_state(param, gaconf):
    new_foo = param['foo'] + random.randint(-9, 9)
    new_param = {**param, 'foo': new_foo}
    return new_param


def example_ga(param_base, gaconf, func, optfunc, input_data, population=10, tolerance=0.1, max_iterations=1000):
    initial_parameters = [generate_param_state(param_base, gaconf) for i in range(population)]

    parameters = initial_parameters
    best_parameters = parameters[0]
    loss = 9999999
    lowest_loss = 999999
    iterations = 0
    loss_values = []

    while loss > tolerance and iterations < max_iterations:
        # Run and get losses
        outputs = [func(params, input_data) for params in parameters]
        losses = [optfunc(o, p) for p, o in zip(parameters, outputs)]
        loss = min(losses)
        loss_values.append(loss)
        lowest_loss = loss if loss < lowest_loss else lowest_loss
        total_loss = sum(losses)

        # Sort parameters
        parameters_index_sorted = list(
            map(lambda x: x[0], sorted(enumerate(losses), key=lambda si: si[1])))

        # Set new best parameters if loss is lowest
        best_parameters = parameters[parameters_index_sorted[0]
                                     ] if loss < lowest_loss else best_parameters

        # Choose next parameter pairs based on
        loss_ratios = [((total_loss - loss)/total_loss)/(population-1) for loss in losses]
        choices_population = [numpy.random.choice(
            population, 2, p=loss_ratios) for i in range(population)]
        new_parameters = [combine_param_state(*[parameters[i] for i in choices])
                          for choices in choices_population]

        mutated_new_parameters = [mutate_param_state(param, gaconf) for param in new_parameters]

        parameters = mutated_new_parameters

        iterations += 1
    last_output = outputs[parameters_index_sorted[0]]
    return best_parameters, iterations, last_output, loss_values


def model(config, data):
    # return config['foo']
    return sum(data) * config['foo']


def loss_function(output, params):
    """The lower the return value the closer it is to optimum."""
    return abs(params['bar'] - output)


example_ga(
    example_param_base, ga_config, model, loss_function, data, max_iterations=5)
# %%
out = model(example_param_base, data)
loss_function(out, example_param_base)


# %%
best_parameters, iterations, last_output, loss_values = example_ga(
    example_param_base, ga_config, model, loss_function, data, max_iterations=5000, tolerance=0.1)
plt.plot(loss_values)
# %%
best_parameters, iterations, last_output

# %%



import random
import numpy as np


def set_seed(seed):
    def wrapper(fn):
        def inner(*args, **kwargs):
            random.seed(seed)
            np.random.seed(seed)
            return fn(*args, **kwargs)
        return inner
    return wrapper



from typing import List
from dataclasses import dataclass

@dataclass
class IterationState:
    """State of current iteration.

    Parameters
    ----------


    parameters: List[dict]
        List of parameters for each element in population
    best_parameters: dict
        Parameters with lowest loss value so far.
    loss: float = 9999999
        The loss from previous run
    lowest_loss: float = 9999999
        The lowest loss so far
    iterations: int = 0
        Number of iterations that have been ran.
    complete: bool = False
        True when iterator meets end conditions

    """
    parameters: List[dict]
    best_parameters: dict
    loss: float = 9999999
    lowest_loss: float = 9999999
    iterations: int = 0
    complete: bool = False




from dataclasses import asdict
from datetime import datetime
from functools import partial
import numpy as np
import random

from paramga.plot import (
    plot_loss,
    plot_observed_compare,
    plot_param,
    plot_param_compare,
)

from .crossover import param_crossover
from .mutation import mutate_param_state
from multiprocessing import Process, Queue
from typing import Any, Callable, Dict, Iterator, List, NamedTuple

from .iteration_state import IterationState

Parameters = dict
ModelOutput = np.ndarray
ModelData = np.ndarray


def get_outputs(func, input_args, input_kwargs):
    return np.array([func(*args, **input_kwargs) for args in input_args])


def get_outputs_parallel(
    func: Callable[[Parameters, ModelData], ModelOutput],
    input_args: List[List[Any]],
    func_kwargs: dict,
    TIMEOUT: float = 100,
) -> ModelOutput:
    """Run the model in parallel

    Parameters
    ----------
    func : Callable[[Parameters, ModelData], ModelOutput]
        model function.
    input_args : List[List[Any]]
        list of dictionaries of arguments for model run
    func_kwargs: dict
        Additional kwargs to pass to model func

    Returns
    -------
    List[ModelOutput]
        The outputs of the Model for each run

    """
    Qu = None
    procs = []
    population=len(input_args)
    try:
        Qu = Queue(maxsize=population)

        def _func(i, args, kwargs):
            # We use a middleware func to link the function outputs to the population index
            try:
                result = func(*args, **kwargs)
            except Exception as e:
                return i, e
            return i, result

        def q_wrap(q, args, kwargs):
            # Add the output of _func to the queue to be collected later
            q.put(_func(*args, kwargs))

        outputs = [None for _ in range(population)]

        # ==== for each run in population we create a process containing a queue writer.
        # NOTE: Args here is (i, model_args)
        for args in enumerate(input_args):
            # Starting Process
            p = Process(target=q_wrap, args=([Qu, args, func_kwargs]))
            procs.append(p)
            p.start()

        # # ==== wait for all processes to complete
        for p in procs:
            p.join(TIMEOUT)

        # # ==== wait for all processes to complete ALT IMPLEMENTATION
        # # Solution from https://stackoverflow.com/questions/26063877/python-multiprocessing-module-join-processes-with-timeout
        # DELAY = .1
        # start = time.time()
        # while time.time() - start <= TIMEOUT:
        #     print("Checking processes")
        #     if not any([p.is_alive() for p in procs]):
        #         # All the processes are done, break now.
        #         break

        #     time.sleep(DELAY)  # Just to avoid hogging the CPU
        # else:
        #     # We only enter this if we didn't 'break' above.
        #     print("timed out, killing all processes")
        #     for p in procs:
        #         p.terminate()
        #         p.join()
        # ===================================================

        # ==== Each process should have written the output of the model to the queue(Qu)
        for p in procs:
            i, res = Qu.get(timeout=.1)
            if isinstance(res, Exception):
                raise res
            outputs[i] = res
            p.close()

    except Exception as e:
        try:
            for p in procs:
                try:
                    print(p.exitcode)
                    p.terminate()
                    p.join(.1)
                except:
                    p.kill()
        except:
            # TODO: Should warn about fail to teardown multi processes
            pass
        print(e)
        raise Exception("Model run multiprocessing failed: {}".format(str(e)))
    return np.array(outputs)


class Logger:
    def __init__(self, log_level):
        self.log_level = log_level

    def log(self, message):
        if self.log_level > 0:
            print(message)


def get_initial_params(
    param_base: dict,
    mutation_conf: dict,
    population: int,
) -> dict:
    return [
        mutate_param_state(param_base, mutation_conf) for _ in range(population)
    ]


def setup_state(
    param_base: dict,
    mutation_conf: dict,
    population: int,
) -> IterationState:
    initial_parameters = get_initial_params(
        param_base,
        mutation_conf,
        population,
    )
    iteration_state = IterationState(
        parameters=initial_parameters,
        best_parameters=initial_parameters[0],
    )
    return iteration_state


def iteration(
    iteration_state: IterationState,
    func: Callable[[Parameters, ModelData], ModelOutput],
    loss_func: Callable[[ModelOutput, Parameters], float],
    population: int,
    mutation_conf: List[dict],
    input_data=None,
    process_outputs: Callable[[np.ndarray], np.ndarray] = None,
    parallel=False,
    func_kwargs: dict=None,
    TIMEOUT: float = 100,
) -> IterationState:
    """Run paramga iteration.

    Parameters
    ----------
    iteration_state : IterationState
        The input iteration state
    func : Callable
        The model function to call with the parameters and data
    loss_func : Callable
        The loss function to run on the model outputs. Should return value between 0-1.
    population : int
        The number of parameter variations to call the model with.
    mutation_conf : List[dict]
        The configuration of parameter mutation.
    input_data : [type], optional
        The input data to pass to the model, by default None
    process_outputs : Callable[[np.ndarray], np.ndarray], optional
        Additional processing of output data before running loss function, by default None
    parallel : bool, optional
        If true will run populations in parallel, by default False
    func_kwargs: dict, optional
        Additional kwargs to pass to model func

    Returns
    -------
    IterationState
        Updated iteration state


    """
    parameters = iteration_state.parameters
    lowest_loss = iteration_state.lowest_loss
    best_parameters = iteration_state.best_parameters
    _func_kwargs = func_kwargs or {}

    # Run and get losses
    input_args = [(params, input_data) for params in parameters]

    run_func = partial(get_outputs_parallel, TIMEOUT=TIMEOUT) if parallel else get_outputs
    outputs = run_func(func, input_args, _func_kwargs)
    outputs_processed = np.array([process_outputs(o)
                                 for o in outputs]) if process_outputs else outputs

    losses = [loss_func(o, p) for p, o in zip(parameters, outputs_processed)]
    curr_min_loss = min(losses)

    # Sort parameters
    parameters_index_sorted = list(
        map(lambda x: x[0], sorted(enumerate(losses), key=lambda si: si[1])))

    # Set new best parameters if loss is lowest
    if curr_min_loss < lowest_loss:
        best_parameters = parameters[parameters_index_sorted[0]]
        lowest_loss = curr_min_loss
    else:
        # If min loss is not lower than lowest loss then add best parameters back to population
        parameters = parameters + [best_parameters]
        losses = losses + [lowest_loss]

    # Choose next parameter pairs using probalistic choice based on loss values
    loss_vals = 1 - (losses - np.min(losses)) / \
        np.ptp(losses) if np.ptp(losses) > 0 else np.ones(len(parameters))
    loss_ratios = loss_vals / sum(loss_vals)

    try:
        choices_population = [np.random.choice(
            len(parameters), 2, p=loss_ratios) for _ in range(population)]
    except ValueError as e:
        print(loss_ratios)
        print(parameters)
        raise e

    # We choose parameter pairs from the population with higher scoring params
    # being more likely to be picked
    crossed_parameters = [param_crossover(*[parameters[i] for i in choices])
                          for choices in choices_population]

    mutated_new_parameters = [mutate_param_state(
        param, mutation_conf) for param in crossed_parameters]

    return IterationState(
        mutated_new_parameters,
        best_parameters,
        curr_min_loss,
        lowest_loss,
        iteration_state.iterations + 1,
    )


def run_iterator(
    param_base: List[Dict],
    mutation_conf: dict,
    func: Callable[[Parameters], float],
    loss_func: Callable[[np.ndarray], float],
    input_data: any,
    process_outputs: Callable[[np.ndarray], np.ndarray] = None,
    population: int = 8,
    tolerance: float = None,
    max_iterations: int = 1000,
    verbose: bool = False,
    parallel: bool = False,
) -> Iterator[IterationState]:
    """Genetic algorithm iterator.

    Yields the next set of params and losses.

    Parameters
    ----------
    param_base : Dict
        The initial parameters for running the model.
    mutation_conf : Dict
        The configuration for mutating the parameters at each step.
    func : Callable[[Parameters, ModelData], ModelOutput]
        The model function to be called with the parameters and data.
    loss_func : Callable[[ModelOutput, Parameters], float]
        The loss function to be called with the model output
    input_data : ModelData
        Input data for the model.
    process_outputs: Callable[[np.ndarray], np.ndarray] = None,
        Function to process outputs before running loss function
    population : int, optional
        The population size for optimization algorithm, by default 8
    tolerance : float, optional
        The tolerance of the optimization algorithm, by default None
    max_iterations : int, optional
        The maximum iterations of the optimization algorithm, by default 1000
    verbose : bool, optional
        Output verbose logs, by default False
    parallel : bool, optional
        Run iterations in parallel, by default False

    Yields
    -------
    IterationState
        The iteration state.

    """
    iteration_state = setup_state(
        param_base,
        mutation_conf,
        population,
    )

    logger = Logger(1 if verbose else 0)

    logger.log("==== Starting ====")
    while iteration_state.iterations < max_iterations and (tolerance is None or iteration_state.loss > tolerance):
        logger.log(
            f"========= Running iteration: {iteration_state.iterations}. Curr loss is {iteration_state.loss}")
        iteration_state = iteration(
            iteration_state,
            func,
            loss_func,
            population,
            mutation_conf,
            input_data=input_data,
            process_outputs=process_outputs,
            parallel=parallel,
        )
        yield iteration_state
    iteration_state.complete = True
    logger.log("===== Complete ======")
    yield iteration_state


def run(
    param_base: List[Dict],
    mutation_conf: dict,
    func: Callable[[Parameters], float],
    loss_func: Callable[[np.ndarray], float],
    input_data: any,
    process_outputs: Callable[[np.ndarray], np.ndarray] = None,
    population: int = 8,
    tolerance: float = None,
    max_iterations: int = 1000,
    verbose: bool = False,
    parallel: bool = False,
) -> IterationState:
    """Wrapper for running the iterator for a set number of iterations.

    Parameters
    ----------
    param_base : Dict
        The initial parameters for running the model.
    mutation_conf : Dict
        The configuration for mutating the parameters at each step.
    func : Callable[[Parameters, ModelData], ModelOutput]
        The model function to be called with the parameters and data.
    loss_func : Callable[[ModelOutput, Parameters], float]
        The loss function to be called with the model output
    input_data : ModelData
        Input data for the model.
    process_outputs : Callable[[ModelOutput], ModelOutput], optional
            Optional function to call on model output before loss function, by default None
    population : int, optional
        [description], by default 8
    tolerance : float, optional
        The tolerance of the optimization algorithm, by default None
    max_iterations : int, optional
        The maximum iterations of the optimization algorithm, by default 1000
    verbose : bool, optional
        Output verbose logs, by default False
    parallel : bool, optional
        Run iterations in parallel, by default False

    Returns
    -------
    IterationState
        IterationState after run.

    """
    for iteration_state in run_iterator(
        param_base,
        mutation_conf,
        func,
        loss_func,
        input_data,
        process_outputs,
        population,
        tolerance,
        max_iterations,
        verbose,
        parallel,
    ):
        if iteration_state.complete:
            return iteration_state
    return iteration_state



class RunStats(NamedTuple):
    total_time: datetime


class Runner:
    """Simple wrapper around run function for class functionality."""

    def __init__(
        self,
        param_base: Dict,
        mutation_conf: Dict,
        func: Callable[[Parameters, ModelData], ModelOutput],
        loss_func: Callable[[ModelOutput, Parameters], float],
        input_data: ModelData,
        func_kwargs: dict = None,
        population: int = 8,
        process_outputs: Callable[[ModelOutput], ModelOutput] = None,
        tolerance: float = None,
        max_iterations: int = 1000,
        verbose: bool = False,
        parallel: bool = False,
        seed: int = 0,
        timeout: float = 100,
    ):
        """Create Runner instance.

        Parameters
        ----------
        param_base : Dict
            The initial parameters for running the model.
        mutation_conf : Dict
            The configuration for mutating the parameters at each step.
        func : Callable[[Parameters, ModelData], ModelOutput]
            The model function to be called with the parameters and data.
        loss_func : Callable[[ModelOutput, Parameters], float]
            The loss function to be called with the model output
        input_data : ModelData
            Input data for the model.
        func_kwargs: dict, optional
            The common kwargs to pass to the model func (Helps with parallel)
        population : int, optional
            The population size for optimization algorithm, by default 8
        process_outputs : Callable[[ModelOutput], ModelOutput], optional
            Optional function to call on model output before loss function, by default None
        tolerance : float, optional
            The tolerance of the optimization algorithm, by default None
        max_iterations : int, optional
            The maximum iterations of the optimization algorithm, by default 1000
        verbose : bool, optional
            Output verbose logs, by default False
        parallel : bool, optional
            Run iterations in parallel, by default False
        seed : int, optional
            The random seed to use, by default 0
        timeout: float
            Timeout for each parallel model run.

        """
        self.param_base = param_base
        self.mutation_conf = mutation_conf
        self.func = func
        self.loss_func = loss_func
        self.input_data = input_data
        self.func_kwargs = func_kwargs
        self.population = population
        self.process_outputs = process_outputs
        self.tolerance = tolerance
        self.max_iterations = max_iterations
        self.verbose = verbose
        self.parallel = parallel
        self.timeout = timeout
        self.seed = seed
        self.stats = RunStats(None)

        self.history = []
        self._store_iterations = False

        self.logger = Logger(1 if verbose else 0)

        self.reset_state()

    def reset_state(self):
        random.seed(self.seed)
        np.random.seed(self.seed)
        self.initial_parameters = get_initial_params(
            self.param_base,
            self.mutation_conf,
            self.population,
        )
        self.iteration_state = setup_state(
            self.param_base,
            self.mutation_conf,
            self.population,
        )
        self.history = [self.iteration_state]
        self.stats = RunStats(None)
        return self

    def __iter__(self):
        def _inner():
            while self.iteration_state.iterations < self.max_iterations \
                    and (self.tolerance is None or self.iteration_state.loss > self.tolerance):
                yield next(self)
        return _inner()

    def __next__(self):
        new_state = iteration(
            iteration_state=self.iteration_state,
            func=self.func,
            loss_func=self.loss_func,
            population=self.population,
            mutation_conf=self.mutation_conf,
            input_data=self.input_data,
            process_outputs=self.process_outputs,
            parallel=self.parallel,
            func_kwargs=self.func_kwargs,
            TIMEOUT=self.timeout,
        )
        self.iteration_state = new_state
        if self._store_iterations:
            self.history.append(new_state)
        return new_state

    def store_iterations(self, v: bool = True):
        self._store_iterations = v
        return self

    def run(self):
        start_time = datetime.now()
        while self.iteration_state.iterations < self.max_iterations \
                and (self.tolerance is None or self.iteration_state.lowest_loss > self.tolerance):
            next(self)
        end_time = datetime.now()
        self.stats = RunStats(end_time - start_time)
        return self

    def get_best_model_output(self):
        """Run the model on the best parameters from the iteration state."""
        return self.func(self.iteration_state.best_parameters, self.input_data, **self.func_kwargs)

    def get_initial_model_output(self):
        """Run the model on the best parameters from the iteration state."""
        return self.func(self.param_base, self.input_data, **self.func_kwargs)

    def iteration_state_dump(self):
        return asdict(self.iteration_state)

    def plot(self, key='loss', ax=None, fig=None):
        if len(self.history) == 0:
            raise ValueError('Must run with store_iteraions on to create plots')

        if key == 'loss':
            loss_values = [s.lowest_loss for s in self.history[1:]]
            return plot_loss(loss_values, ax=ax, fig=fig)
        else:
            raise ValueError(f'{key} is invalid key')

    def plot_param(self, key, ax=None, fig=None):
        if len(self.history) == 0:
            raise ValueError('Must run with store_iteraions on to create plots')

        param_values = np.array(
            # [[p[key] for p in self.initial_parameters]] +
            [[p[key] for p in s.parameters] for s in self.history[:-1]])
        best_values = np.array([s.best_parameters[key] for s in self.history[1:]])
        loss_values = [s.lowest_loss for s in self.history[1:]]
        return plot_param(key, param_values, best_values, loss_values, ax=ax, fig=fig)

    def plot_param_compare(self, key_a, key_b, ax=None, fig=None):
        best_values_a = np.array([s.best_parameters[key_a] for s in self.history[1:]])
        best_values_b = np.array([s.best_parameters[key_b] for s in self.history[1:]])
        loss_values = [s.loss for s in self.history[1:]]
        return plot_param_compare(key_a, key_b, best_values_a, best_values_b, loss_values, ax=ax, fig=fig)

    def plot_compare_with_observed(
        self, observed_values, ax=None, fig=None,
    ):
        values = self.func(self.iteration_state.best_parameters, self.input_data, **self.func_kwargs)
        try:
            assert type(observed_values) == np.ndarray
            assert type(values) == np.ndarray
        except:
            raise AssertionError('Observed and model values must by numpy arrays.')
        return plot_observed_compare(
            observed_values,
            values,
            ax=ax,
            fig=fig,
            xlim=(min(observed_values), max(observed_values)),
            ylim=(min(observed_values), max(observed_values)),
        )



import click


@click.group()
def cli():
    """Main cli entrypoint."""
    click.echo("Welcome to paramga")


@cli.command()
def run(

):

    click.echo("Complete!")


if __name__ == "__main__":
    ''' Get the config and input data locations from the cli
    Then loads these into the config and runs the model

    '''
    cli()



import random
from .types import Parameters

def mutate_param_state(
    param: Parameters,
    gaconf: dict,
) -> Parameters:
    """Mutate the parameter state.

    Parameters
    ----------
    param : dict
        input parameters
    gaconf : dict
        dictionary of mutations

    Returns
    -------
    dict
        Mutated parameters

    """
    param_copy = {**param}
    for k, var in gaconf.items():
        if var['type'] == 'number':
            step = var['step']
            r = random.randrange(-step, step)
        elif var['type'] == 'float':
            r = 2 * (random.random()-0.5) * var['step']
        else:
            raise ValueError('Invalid var type')
        if r == 0:
            continue

        new_val = param[k] + r
        new_val = min(new_val, var['max']) if var['max'] is not None else new_val
        new_val = max(new_val, var['min']) if var['min'] is not None else new_val
        param_copy[k] = new_val
    return param_copy



import pickle
import numpy as np

from paramga.run import Runner


def demo_state():
    return {
        "foo": 10,
        "bar": 360,
    }

def demo_conf():
    return {
        "foo": {
            "type": "number",
            "min": 1,
            "max": 80,
            "step": 8,
        }
    }

def model(params, data):
    return np.array([d * params['foo'] for d in data])
def demo_model():
    def model(params, data):
        return np.array([d * params['foo'] for d in data])
    return model

def process_outputs(outputs):
    return outputs + 10

def demo_process_outputs():
    def process_outputs(outputs):
        return outputs + 10
    return process_outputs

def loss_function(output, params):
    return abs(params['bar'] - sum(output) + 40)

def demo_loss_function():
    def loss_function(output, params):
        return abs(params['bar'] - sum(output) + 40)
    return loss_function

def demo_data():
    return [1, 2, 3, 3]

def demo_best_params():
    return {
        "foo": 40,
        "bar": 360,
    }

class TestRunner:



    def _default_instance(self, **kwargs):
        default_args = dict(
            param_base=demo_state(),
            mutation_conf=demo_conf(),
            func=model,
            loss_func=loss_function,
            input_data=demo_data(),
            process_outputs=process_outputs,
            population=8,
            max_iterations=100,
            tolerance=0.05,
        )
        return Runner(**{**default_args, **kwargs})

    def test_can_pickle_runner_state(self):
        runner = self._default_instance()
        pickle.dumps(runner.iteration_state)


    def test_can_pickle_runner(self):
        runner = self._default_instance()
        pickle.dumps(runner)


from paramga.mutation import mutate_param_state
from paramga.random_helpers import set_seed


class TestMutateParamState:

    def demo_state(self):
        return {
            "foo": 10,
            "bar": 100,
        }

    def demo_conf(self):
        return {
            "foo": {
                "type": "number",
                "min": 3,
                "max": 14,
                "step": 1,
            }
        }

    def test_with_simple_config(self):
        out_state = set_seed(1)(mutate_param_state)(self.demo_state(), self.demo_conf())
        assert out_state['foo'] == 9

    def test_moves_by_step(self):
        input_state = self.demo_state()
        out_state = set_seed(1)(mutate_param_state)(input_state, self.demo_conf())
        assert out_state['foo'] == input_state['foo'] - 1

    def test_is_limited_by_max(self):
        input_state = self.demo_state()
        input_state['foo'] = 99
        out_state = set_seed(1)(mutate_param_state)(input_state, self.demo_conf())
        assert out_state['foo'] == 14



"""Helpers for plotting genetic algorithm outputs."""

from typing import List, Tuple
from matplotlib import pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable


def plot_loss(loss_values, ax=None, fig=None):
    _fig = fig if fig is not None else plt.figure()
    _ax = ax if ax is not None else plt.axes()

    x = list(range(len(loss_values)))
    _ax.scatter(x, loss_values)
    return _fig, _ax


def plot_param(
    param: str,
    values: np.ndarray,
    best_values: np.ndarray,
    loss_values: np.ndarray,
    ax: plt.Axes = None,
    fig: plt.Figure = None,
    vmin: int = 0,
    vmax: int = 1,
) -> Tuple[plt.Figure, plt.Axes]:
    _fig = fig if fig is not None else plt.figure()
    _ax = ax if ax is not None else plt.axes()

    population = len(values[0])
    iterations = len(values)
    x = list(range(iterations))
    for y in range(population):
        _ax.scatter(x, values[:, y], c='grey', s=2)

    _ax.plot(x, best_values, c='red', zorder=1)
    best_scatter = _ax.scatter(x, best_values, c=loss_values, vmin=vmin, vmax=vmax, zorder=2)

    _ax.set_ylabel(param)
    _ax.set_xlabel('iterations')

    divider = make_axes_locatable(_ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)
    _fig.colorbar(best_scatter, cax=cax, ax=_ax)
    return _fig, _ax


def plot_param_compare(
    param_a: str,
    param_b: str,
    values_a: List[float],
    values_b: List[float],
    loss_values: List[float],
    ax: plt.Axes = None,
    fig: plt.Figure = None,
    vmin: float = 0,
    vmax: float = 1,
    xlim: Tuple[float, float] = (0, 300),
    ylim: Tuple[float, float] = (0, 300),
) -> Tuple[plt.Figure, plt.Axes]:
    _fig = fig if fig is not None else plt.figure()
    _ax = ax if ax is not None else plt.axes()

    scatter = _ax.scatter(values_a, values_b, c=loss_values, vmin=vmin, vmax=vmax, zorder=1)

    _ax.set_xlabel(param_a)
    _ax.set_ylabel(param_b)

    _ax.set_xlim(xlim)
    _ax.set_ylim(ylim)

    divider = make_axes_locatable(_ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)
    _fig.colorbar(scatter, cax=cax, ax=_ax)
    return _fig, _ax


def plot_observed_compare(
    observed_values: List[float],
    modelled_values: List[float],
    ax: plt.Axes = None,
    fig: plt.Figure = None,
    vmin: float = 0,
    vmax: float = 1,
    xlim: Tuple[float, float] = (0, 300),
    ylim: Tuple[float, float] = (0, 300),
) -> Tuple[plt.Figure, plt.Axes]:
    _fig = fig if fig is not None else plt.figure()
    _ax = ax if ax is not None else plt.axes()

    scatter = _ax.scatter(observed_values, modelled_values, vmin=vmin, vmax=vmax, zorder=1)

    _ax.set_xlim(*xlim)
    _ax.set_ylim(*ylim)
    _ax.plot([xlim[0], xlim[1]], [ylim[0], ylim[1]])

    _ax.set_xlabel('observed')
    _ax.set_ylabel('modelled')

    # divider = make_axes_locatable(_ax)
    # cax = divider.append_axes('right', size='5%', pad=0.05)
    # _fig.colorbar(scatter, cax=cax, ax=_ax)
    return _fig, _ax



"""Example of using the library."""
# %%

from paramga.run import run_iterator
from matplotlib import pyplot as plt
from math import exp
import numpy as np
# %%


def demo_state():
    return {
        "foo": 800,
        "bar": 360,
        "zoo": 400
    }

def demo_conf():
    return {
        "foo": {
            "type": "number",
            "min": 1,
            "max": 1000,
            "step": 10,
        },
        "zoo": {
            "type": "number",
            "min": 80,
            "max": 500,
            "step": 10,
        }
    }

def demo_model():
    def model(params, data):
        return sum(data) * params['foo']
    return model

def demo_loss_function():
    def loss_function(output, params):
        x = params['bar'] - output
        return abs((x / (1 + exp(x))))
        # return abs(params['bar'] - output)
    return loss_function

def demo_data():
    return [1, 2, 3, 3]

def demo_best_params():
    return {
        "foo": 40,
        "bar": 360,
        "zoo": 10,
    }


fg, axs = plt.subplots(nrows=3, figsize=(20,5))
def demo_model_b(params, data):
    return (params['foo'] - params['zoo'])** 2 if params['foo'] < 100 else params['foo']**2 #sum([v for v in params.values()])
for i in range(10):
    model_iterator = run_iterator(
        demo_state(),
        demo_conf(),
        demo_model_b,
        demo_loss_function(),
        demo_data(),
        max_iterations=100,
        tolerance=None,
        verbose=False,
        population=(i + 1) * 3,
    )

    states = list(model_iterator)
    states

    loss = [s.loss for s in states]
    axs[0].plot(loss, label=i)

    param_foo = [s.parameters[0]['foo'] for s in states]
    axs[1].plot(param_foo, label=i)


    param_bar = [s.parameters[0]['zoo'] for s in states]
    axs[2].plot(param_bar, label=i)
    print(states[0].best_parameters)
for ax in axs:
    ax.legend()
# %%
states[-1].best_parameters

# %%
# Plot
for i in range(5):
    model_iterator = run_iterator(
        demo_state(),
        demo_conf(),
        demo_model_b,
        demo_loss_function(),
        demo_data(),
        max_iterations=200,
        tolerance=0.1,
        verbose=False,
        population=(i + 1) * 3,
    )

    states = list(model_iterator)
    states

    loss = [s.loss for s in states]
    axs[0].plot(loss, label=i)

    param_foo = [s.parameters[0]['foo'] for s in states]
    axs[1].plot(param_foo, label=i)


    param_bar = [s.parameters[0]['zoo'] for s in states]
    axs[2].plot(param_bar, label=i)
    print(states[0].best_parameters)
    x = [s.best_parameters['foo'] for s in states]
    y = [s.best_parameters['zoo'] for s in states]
    losses = np.array([s.loss for s in states])
    plt.scatter(x, y, c=1/losses, s=1/losses)
    # plt.plot(x, y)
plt.colorbar()
# %%


# %%



import setuptools
import pathlib

here = pathlib.Path(__file__).parent.resolve()

with open("README.md", "r") as fh:
    long_description = fh.read()

long_description = (here / 'README.md').read_text(encoding='utf-8')

setuptools.setup(
    name="paramga",
    version="0.4.8",
    author="SBland",
    author_email="sblandcouk@gmail.com",
    description="Parameter Regression GA tool",
    long_description=long_description,
    python_requires='>=3.6',
    setup_requires=[
        'pytest-cov',
        'pytest-runner',
        'snapshottest'
    ],
    tests_require=['pytest', 'numpy'],
    extras_require={'test': ['pytest', 'numpy']},
    install_requires=['numpy'],
    packages=setuptools.find_packages(),
    package_dir={'paramga': 'paramga'},
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
        'License :: OSI Approved :: MIT License',
    ],
    keywords='genetic algorithm, optimization, scientific modelling',  # Optional
)



import pytest
import numpy as np
from unittest.mock import MagicMock
from paramga.random_helpers import set_seed
from paramga.run import IterationState, Runner, get_outputs_parallel, iteration, run, run_iterator


__module_loc__ = 'paramga.run'


class TestIteration:

    class TestUnitTests:

        @pytest.fixture(autouse=True)
        def _setup(self, mocker):
            # default args
            self.population = 10
            self.parameters = [MagicMock() for _ in range(self.population)]
            self.iteration_state = IterationState(
                self.parameters,
                self.parameters,
                loss=999999,
                lowest_loss=99999,
            )
            self.mock_func_result = 99
            self.mock_func_result_processed = 99.9
            self.func = MagicMock(return_value=self.mock_func_result)
            self.loss_func = MagicMock(side_effect=lambda out,
                                       param: 0 if param == self.parameters[0] else 1)
            self.mutation_conf = {}
            self.input_data = []
            self.parallel = False

            # Mocks
            self.mutated_params = MagicMock()
            self.crossover_params = MagicMock()
            self.mock_output_postprocess = MagicMock(return_value=self.mock_func_result_processed)

            self.mock_mutate_param_state = mocker.patch(
                __module_loc__ + '.mutate_param_state', return_value=self.mutated_params)

            self.mock_param_crossover = mocker.patch(
                __module_loc__ + '.param_crossover', return_value=self.crossover_params)

        def _default_run(self, **kwargs):
            default_args = dict(
                iteration_state=self.iteration_state,
                func=self.func,
                loss_func=self.loss_func,
                population=self.population,
                mutation_conf=self.mutation_conf,
                input_data=self.input_data,
                process_outputs=self.mock_output_postprocess,
                parallel=self.parallel,
            )
            _kwargs = {
                **default_args,
                **kwargs,
            }
            return set_seed(1)(iteration)(**_kwargs)

        def test_runs_ok(self):
            self._default_run()

        def test_model_func_is_called(self):
            self._default_run()
            assert self.func.call_count == self.population
            for params in self.parameters:
                self.func.assert_any_call(
                    params,
                    self.input_data,
                )

        def test_output_post_process_called(self):
            self._default_run()
            assert self.mock_output_postprocess.call_count == self.population
            self.mock_output_postprocess.assert_any_call(
                self.mock_func_result,
            )

        def test_loss_func_is_called(self):
            self._default_run()
            assert self.loss_func.call_count == self.population
            self.loss_func.assert_called_with(
                self.mock_func_result_processed,
                self.parameters[-1],
            )

        def test_param_crossover_called(self):
            self._default_run()
            assert self.mock_param_crossover.call_count == self.population
            self.mock_param_crossover.assert_any_call(
                self.parameters[0],
                self.parameters[0],
            )

        def test_should_call_mutate_param_state(self):
            self._default_run()
            assert self.mock_mutate_param_state.call_count == self.population

        def test_returns_updated_state(self):
            out = self._default_run()
            assert out.iterations == self.iteration_state.iterations + 1
            assert out.parameters == [self.mutated_params for _ in range(self.population)]

        def test_should_pass_extra_model_args_to_model_run(self):
            self.func_kwargs = {
                "a": 123,
            }
            self._default_run(func_kwargs=self.func_kwargs)
            assert self.func.call_count == self.population
            for params in self.parameters:
                self.func.assert_any_call(
                    params,
                    self.input_data,
                    **self.func_kwargs,
                )

    class TestFuncTests:

        @pytest.fixture(autouse=True)
        def _setup(self, mocker):
            # default args
            self.population = 10
            self.parameters = [{"foo": 1, "bar": 1} for _ in range(self.population)]
            self.iteration_state = IterationState(
                self.parameters,
                self.parameters,
                loss=999999,
                lowest_loss=99999,
            )
            self.func = lambda params, data: params['foo'] * 2 + params['bar'] * 4
            self.loss_func = lambda output, params: output - 42
            self.max_foo = 10
            self.mutation_conf = {
                "foo": {
                    "type": "float",
                    "min": 1,
                    "max": self.max_foo,
                    "step": 0.1,
                },
                "bar": {
                    "type": "number",
                    "min": 1,
                    "max": 20,
                    "step": 1,
                },
            }
            self.input_data = []
            self.parallel = False

        def _default_run(self, **kwargs):
            default_args = dict(
                iteration_state=self.iteration_state,
                func=self.func,
                loss_func=self.loss_func,
                population=self.population,
                mutation_conf=self.mutation_conf,
                input_data=self.input_data,
                parallel=self.parallel,
            )
            _kwargs = {
                **default_args,
                **kwargs,
            }
            return set_seed(1)(iteration)(**_kwargs)

        def test_run_without_error(self):
            self._default_run()

        def test_should_have_reduced_loss(self):
            out = self._default_run()
            assert out.loss < self.iteration_state.loss

        def test_should_have_mutated_parameters(self):
            out = self._default_run()
            assert len(out.parameters) == len(self.parameters)
            assert out.parameters != self.parameters
            assert all(o['foo'] < self.max_foo for o in out.parameters)

        def test_can_handle_different_loss_functions(self):
            self._default_run(loss_func=lambda *args, **kwargs: 99)
            self._default_run(loss_func=lambda *args, **kwargs: -100)


class TestRun:

    def demo_state(self):
        return {
            "foo": 10,
            "bar": 360,
        }

    def demo_conf(self):
        return {
            "foo": {
                "type": "number",
                "min": 1,
                "max": 80,
                "step": 8,
            }
        }

    def demo_model(self):
        def model(params, data):
            return sum(data) * params['foo']
        return model

    def demo_process_outputs_func(self):
        def process_outputs(outputs):
            return outputs + 100
        return process_outputs

    def demo_loss_function(self):
        def loss_function(output, params):
            return abs(params['bar'] - output)
        return loss_function

    def demo_data(self):
        return [1, 2, 3, 3]

    def demo_best_params(self):
        return {
            "foo": 40,
            "bar": 360,
        }

    def _default_run(self, **kwargs):
        default_args = dict(
            param_base=self.demo_state(),
            mutation_conf=self.demo_conf(),
            func=self.demo_model(),
            loss_func=self.demo_loss_function(),
            input_data=self.demo_data(),
            population=8,
            tolerance=0.05,
            max_iterations=100,
            verbose=False,
            parallel=False,
        )
        _kwargs = {**default_args, **kwargs}
        return set_seed(1)(run)(**_kwargs)

    def test_simple_run(self):
        iteration_state = self._default_run()
        assert 100 > iteration_state.iterations > 5
        assert iteration_state.best_parameters == self.demo_best_params()

    def test_limited_by_max_iterations(self):
        tolerance = 0.00000001
        iteration_state = self._default_run(max_iterations=5, tolerance=tolerance)
        assert iteration_state.loss > tolerance
        assert iteration_state.iterations == 5
        assert iteration_state.best_parameters != self.demo_best_params()


    def test_running_in_parallel(self):
        tolerance = 0.0001
        iteration_state = self._default_run(
            tolerance=tolerance,
            parallel=True,
        )

        assert 100 > iteration_state.iterations > 5
        assert iteration_state.loss < tolerance
        assert iteration_state.best_parameters == self.demo_best_params()


class TestRunCls:

    def demo_state(self):
        return {
            "foo": 10,
            "bar": 360,
        }

    def demo_conf(self):
        return {
            "foo": {
                "type": "number",
                "min": 1,
                "max": 80,
                "step": 8,
            }
        }

    def demo_model(self):
        def model(params, data, abc=None):
            assert abc # example kwarg
            return np.array([d * params['foo'] for d in data])
        return model

    def demo_process_outputs(self):
        def process_outputs(outputs):
            return outputs + 10
        return process_outputs

    def demo_loss_function(self):
        def loss_function(output, params):
            return abs(params['bar'] - sum(output) + 40)
        return loss_function

    def demo_data(self):
        return [1, 2, 3, 3]

    def demo_best_params(self):
        return {
            "foo": 40,
            "bar": 360,
        }
    def demo_kwargs(self):
        return dict(abc=123)

    def _default_instance(self, **kwargs):
        default_args = dict(
            param_base=self.demo_state(),
            mutation_conf=self.demo_conf(),
            func=self.demo_model(),
            loss_func=self.demo_loss_function(),
            input_data=self.demo_data(),
            process_outputs=self.demo_process_outputs(),
            func_kwargs=self.demo_kwargs(),
            population=8,
            max_iterations=100,
            tolerance=0.05,
        )
        return Runner(**{**default_args, **kwargs})

    def test_simple_run(self):
        runner = self._default_instance()
        iteration_state = runner.run().iteration_state
        assert 100 > iteration_state.iterations > 5
        assert iteration_state.best_parameters == self.demo_best_params()

    def test_limited_by_max_iterations(self):
        tolerance = 0.00000001

        runner = self._default_instance(
            tolerance=tolerance,
            max_iterations=5,
        )

        iteration_state = runner.run().iteration_state
        assert iteration_state.loss > tolerance
        assert iteration_state.iterations == 5
        assert iteration_state.best_parameters != self.demo_best_params()

    def test_best_parameters_is_best_from_current_set(self):
        tolerance = 0.00000001
        runner = self._default_instance(
            tolerance=tolerance,
            max_iterations=5,
        )

        iteration_state = runner.run().iteration_state
        assert iteration_state.loss > tolerance
        assert iteration_state.iterations == 5
        assert iteration_state.best_parameters != self.demo_best_params()


    def test_running_in_parallel(self):
        tolerance = 0.00000001
        runner = self._default_instance(
            tolerance=tolerance,
            parallel=True,
        )

        iteration_state = runner.run().iteration_state
        assert 100 > iteration_state.iterations > 5
        assert iteration_state.loss <= tolerance
        assert iteration_state.loss == 0
        assert iteration_state.best_parameters == self.demo_best_params()


    def test_run_as_iterator(self):
        tolerance = 0.05
        runner = self._default_instance(
            tolerance=tolerance,
            parallel=True,
        )

        for iteration_state in iter(runner):
            pass
        assert 100 > iteration_state.iterations > 5
        assert iteration_state.loss <= tolerance


    def test_can_store_iteration_data(self):
        tolerance = 0.05
        runner = self._default_instance(
            tolerance=tolerance,
            parallel=True,
        )
        runner.store_iterations()
        final_state = runner.run().iteration_state
        assert len(runner.history) == final_state.iterations + 1

    # def test_should_store_parameters_in_order(self):
    #     tolerance = 0.05
    #     runner = Runner(
    #         self.demo_state(),
    #         self.demo_conf(),
    #         self.demo_model(),
    #         self.demo_loss_function(),
    #         self.demo_data(),
    #         max_iterations=100,
    #         tolerance=tolerance,
    #         parallel=True,
    #     )
    #     runner.store_iterations()
    #     runner.run().iteration_state
    #     assert runner.initial_parameters == runner.history[0].parameters


    def test_should_plot_loss(self):
        tolerance = 0.001
        runner = self._default_instance(
            tolerance=tolerance,
            parallel=True,
        )
        runner = self._default_instance(
            tolerance=tolerance,
        )
        runner.store_iterations()
        runner.run()
        runner.plot()

    def test_should_plot_param(self):
        tolerance = 0.001
        runner = self._default_instance(
            tolerance=tolerance,
        )
        runner.store_iterations()
        runner.run()
        runner.plot_param('foo')

    def test_should_plot_param_compare(self):
        tolerance = 0.001
        runner = self._default_instance(
            tolerance=tolerance,
        )
        runner.store_iterations()
        runner.run()
        runner.plot_param_compare('foo', 'bar')

    def test_should_plot_observed_compare(self):
        tolerance = 0.001
        runner = self._default_instance(
            tolerance=tolerance,
        )
        runner.store_iterations()
        runner.run()
        print(runner.func(runner.iteration_state.best_parameters, runner.input_data, **self.demo_kwargs()))
        observed_data = np.array(runner.input_data)
        runner.plot_compare_with_observed(observed_data)

    def test_setting_seed_makes_results_reproducable(self):
        runner = self._default_instance(seed=0, max_iterations=3)
        runner.store_iterations()
        runner.run()
        assert runner.iteration_state.lowest_loss == 198
        runner = self._default_instance(seed=1, max_iterations=3)
        runner.store_iterations()
        runner.run()
        assert runner.iteration_state.lowest_loss != 198
        assert runner.iteration_state.lowest_loss == 108


    def test_setting_seed_makes_results_reproducable_parallel(self):
        runner = self._default_instance(seed=0, max_iterations=3, parallel=True)
        runner.store_iterations()
        runner.run()
        assert runner.iteration_state.lowest_loss == 198
        runner = self._default_instance(seed=1, max_iterations=3)
        runner.store_iterations()
        runner.run()
        assert runner.iteration_state.lowest_loss != 198
        assert runner.iteration_state.lowest_loss == 108

    def test_can_get_initial_output(self):
        tolerance = 0.00000001
        runner = self._default_instance(
            tolerance=tolerance,
            max_iterations=100,
        )
        runner.store_iterations()
        runner.run()

        initial_output = runner.get_initial_model_output()
        expected_initial_output = self.demo_model()(self.demo_state(), self.demo_data(), **self.demo_kwargs())
        assert len(initial_output) == len(expected_initial_output)
        for a, b in zip(initial_output, expected_initial_output):
            assert a == b

    def test_can_get_best_output(self):
        tolerance = 0.00000001
        runner = self._default_instance(
            tolerance=tolerance,
            max_iterations=100,
        )
        runner.store_iterations()
        runner.run()

        best_output = runner.get_best_model_output()
        expected_best_output = self.demo_model()(self.demo_best_params(), self.demo_data(), **self.demo_kwargs())
        assert len(best_output) == len(expected_best_output)
        for a, b in zip(best_output, expected_best_output):
            assert a == b

class TestRunParallelIterator:

    def demo_state(self):
        return {
            "foo": 10,
            "bar": 360,
        }

    def demo_conf(self):
        return {
            "foo": {
                "type": "number",
                "min": 3,
                "max": 80,
                "step": 8,
            }
        }

    def demo_loss_function(self):
        def loss_function(output, params):
            return abs(params['bar'] - output)
        return loss_function

    def demo_data(self):
        return [1, 2, 3, 3]

    def demo_best_params(self):
        return {
            "foo": 40,
            "bar": 360,
        }

    @set_seed(1)
    def test_simple_run(self):
        def demo_model_b(params, data):
            return sum(data) * params['foo']

        model = run_iterator(
            self.demo_state(),
            self.demo_conf(),
            demo_model_b,
            self.demo_loss_function(),
            self.demo_data(),
            max_iterations=100,
            tolerance=0.005,
            verbose=True,
        )
        iterations_state = next(model)
        assert iterations_state.iterations == 1
        assert iterations_state.loss < 99999

class TestGetOutputsParallel:
    def test_get_outputs_parallel(self):
        from time import sleep
        input_args = [
            [0, 0.1],
            [1, 0.4],
            [2, 0.3],
            [3, 0.4],
        ]

        def func(i, t, j=0, k=1):
            sleep(t)
            return i, t, j+k
        func_kwargs = { "j": 98, "k": 1}
        # func_kwargs = {}
        outputs = get_outputs_parallel(func, input_args, func_kwargs)
        expected_outputs = [
            [i, t, 99] for i, t in input_args
        ]
        np.testing.assert_array_equal(outputs, expected_outputs)

    def test_should_catch_errors(self):
        from time import sleep
        input_args = [
            [0, 3],
            [1, 3],
            [2, 3],
            [3, 3],
        ]

        def func(i, t, j=0, k=1):
            sleep(t)
            raise Exception("RANDOM_ERROR_IN_PROCESS")

        func_kwargs = {}
        with pytest.raises(Exception) as e:
            get_outputs_parallel(func, input_args, func_kwargs, TIMEOUT=1)

        assert "Model run multiprocessing failed" in str(e)
        assert "RANDOM_ERROR_IN_PROCESS" in str(e)
        # assert "Model run multiprocessing failed" in str(e)